use std::{hint::cold_path, sync::Arc};

use crate::{
    context::Context,
    job::{QueuedJob as Job, StackJob},
    latch::NopLatch,
    workerthread::WorkerThread,
};

impl WorkerThread {
    #[inline]
    #[tracing::instrument(level = "trace", skip_all)]
    fn join_seq<A, B, RA, RB>(&self, a: A, b: B) -> (RA, RB)
    where
        A: FnOnce() -> RA,
        B: FnOnce() -> RB,
    {
        let span = tracing::trace_span!("join_seq");
        let _guard = span.enter();

        let rb = b();
        let ra = a();

        (ra, rb)
    }

    #[inline]
    pub(crate) fn join_heartbeat_every<A, B, RA, RB>(&self, a: A, b: B) -> (RA, RB)
    where
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB,
        RA: Send,
    {
        // self.join_heartbeat_every_inner::<A, B, RA, RB, 2>(a, b)
        self.join_heartbeat(a, b)
    }

    /// This function must be called from a worker thread.
    #[inline(always)]
    fn join_heartbeat_every_inner<A, B, RA, RB, const TIMES: usize>(&self, a: A, b: B) -> (RA, RB)
    where
        RA: Send,
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB,
    {
        // SAFETY: each worker is only ever used by one thread, so this is safe.
        let count = self.join_count.get();
        let queue_len = unsafe { self.queue.as_ref_unchecked().len() };
        self.join_count.set(count.wrapping_add(1) % TIMES as u8);

        // TODO: add counter to job queue, check for low job count to decide whether to use heartbeat or seq.
        // see: chili

        // SAFETY: this function runs in a worker thread, so we can access the queue safely.
        if count == 0 || queue_len < 3 {
            cold_path();
            self.join_heartbeat(a, b)
        } else {
            self.join_seq(a, b)
        }
    }

    /// This function must be called from a worker thread.
    #[inline]
    #[tracing::instrument(level = "trace", skip_all)]
    fn join_heartbeat<A, B, RA, RB>(&self, a: A, b: B) -> (RA, RB)
    where
        RA: Send,
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB,
    {
        use std::panic::{AssertUnwindSafe, catch_unwind, resume_unwind};

        let a = StackJob::new(a, NopLatch);
        let job = Job::from_stackjob(&a, self.heartbeat.raw_latch());

        self.push_back(&job);

        self.tick();

        let rb = match catch_unwind(AssertUnwindSafe(|| b())) {
            Ok(val) => val,
            Err(payload) => {
                tracing::debug!("join_heartbeat: b panicked, waiting for a to finish");
                cold_path();
                // if b panicked, we need to wait for a to finish
                self.wait_until_latch(&job);
                resume_unwind(payload);
            }
        };

        let ra = if !job.is_shared() {
            tracing::trace!("join_heartbeat: job is not shared, running a() inline");
            // we pushed the job to the back of the queue, any `join`s called by `b` on this worker thread will have already popped their job, or seen it be executed.
            self.pop_back();

            // a is allowed to panic here, because we already finished b.
            unsafe { a.unwrap()() }
        } else {
            match self.wait_until_queued_job(&job) {
                Some(t) => crate::util::unwrap_or_panic(t),
                None => {
                    tracing::trace!(
                        "join_heartbeat: job was shared, but reclaimed, running a() inline"
                    );
                    // the job was shared, but not yet stolen, so we get to run the
                    // job inline
                    unsafe { a.unwrap()() }
                }
            }
        };

        (ra, rb)
    }
}

impl Context {
    #[inline]
    pub fn join<A, B, RA, RB>(self: &Arc<Self>, a: A, b: B) -> (RA, RB)
    where
        A: FnOnce() -> RA + Send,
        B: FnOnce() -> RB + Send,
        RA: Send,
        RB: Send,
    {
        // SAFETY: join_heartbeat_every is safe to call from a worker thread.
        self.run_in_worker(move |worker| worker.join_heartbeat_every::<_, _, _, _>(a, b))
    }
}

/// run two closures potentially in parallel, in the global threadpool.
#[allow(dead_code)]
pub fn join<A, B, RA, RB>(a: A, b: B) -> (RA, RB)
where
    A: FnOnce() -> RA + Send,
    B: FnOnce() -> RB + Send,
    RA: Send,
    RB: Send,
{
    join_in(Context::global_context().clone(), a, b)
}

/// run two closures potentially in parallel, in the global threadpool.
#[allow(dead_code)]
fn join_in<A, B, RA, RB>(context: Arc<Context>, a: A, b: B) -> (RA, RB)
where
    A: FnOnce() -> RA + Send,
    B: FnOnce() -> RB + Send,
    RA: Send,
    RB: Send,
{
    context.join(a, b)
}
